// This file has been prepared for Doxygen automatic documentation generation.
/*! \file ********************************************************************
*
* Atmel Corporation
*
* - File              : sensor_three_phase_BLDC.c
* - Compiler          : IAR EWAAVR 2.28a/3.10c
* - Support mail      : avr@atmel.com
* - Supported devices : All devices with 3 PWM channels and Pin
*                       Change Interrupts can be used.
*                       The example is written for ATmega48.
* - AppNote           : AVR443 - Sensor based control of three phase
*                       brushless DC motor
* - Description       : Example of how to use an ATmega48 to control
*                       a brushless DC motor using Hall sensor for rotor
*                       position feedback.
*
* $Revision: 1.1 $
* $Date: Wednesday, June 15, 2005 07:38:42 UTC $
*****************************************************************************/
#include <ioavr.h>
#include <inavr.h>
#include "sensor_three_phase_BLDC.h"



/*! \brief CCW rotation patterns.
 *
 * Configuration of pin drive levels
 * and timer COM bits in different
 * phases for CounterClockWise rotation.
 */
unsigned char drvPatternsCCW[] = {
  0,    //Stop
  // MC_PORT drive config
  PDP2_CCW, //Phase2
  PDP4_CCW, //Phase4
  PDP3_CCW, //Phase3
  PDP6_CCW, //Phase6
  PDP1_CCW, //Phase1
  PDP5_CCW, //Phase5
  // Configuration of Output Compare operation for timer 0
  COM0P2_CCW, //Phase2
  COM0P4_CCW, //Phase4
  COM0P3_CCW, //Phase3
  COM0P6_CCW, //Phase6
  COM0P1_CCW, //Phase1
  COM0P5_CCW, //Phase5
  // Configuration of Output Compare operation for timer 2
  COM2P2_CCW, //Phase2
  COM2P4_CCW, //Phase4
  COM2P3_CCW, //Phase3
  COM2P6_CCW, //Phase6
  COM2P1_CCW, //Phase1
  COM2P5_CCW  //Phase5
};



/*! \brief CW rotation patterns.
 *
 * Configuration of pin drive levels
 * and timer COM bits in different
 * phases for ClockWise rotation.
 */
unsigned char drvPatternsCW[] = {
  0,    //Stop
  // MC_PORT drive config
  PDP2_CW, //Phase2
  PDP6_CW, //Phase6
  PDP1_CW, //Phase1
  PDP4_CW, //Phase4
  PDP3_CW, //Phase3
  PDP5_CW,  //Phase5
  // Configuration of Output Compare operation for timer 0
  COM0P2_CW, //Phase2
  COM0P6_CW, //Phase6
  COM0P1_CW, //Phase1
  COM0P4_CW, //Phase4
  COM0P3_CW, //Phase3
  COM0P5_CW,  //Phase5
  // Configuration of Output Compare operation for timer 2
  COM2P2_CW, //Phase2
  COM2P6_CW, //Phase6
  COM2P1_CW, //Phase1
  COM2P4_CW, //Phase4
  COM2P3_CW, //Phase3
  COM2P5_CW  //Phase5
};

								

//! Stores the current motor driver pattern.
__regvar __no_init unsigned char *pDrvPattern @14;
//__regvar __no_init unsigned char *pComPattern @12;



//! Used for optimized temporary varables.
__regvar __no_init union _fastTemp{
  unsigned int word;
  struct{
    unsigned char LByte;
    unsigned char HByte;  //Hbyte = Zero
  };
} fastTemp @12;



__regvar __no_init unsigned char hallMask @11; //!< Workaround for internal compiler error
__regvar __no_init unsigned char count @10; //!< Optimized variable decremented every pin change int.



#pragma vector = PCINT0_vect
//! Pin Change Interrupt Service Routine.
__interrupt void PCINT0_ISR( void )
{
  unsigned char *pTemp;
  fastTemp.word = ((PIN_HALL & hallMask)>>1);  // Read Hall, Mask Pins, shift to use as pointer offset
//  Line below is desirable performance wise, but causes an internal error in compiler
//  fastTemp.LByte = (PIN_HALL & HALL_MASK)>>1;   // Read Hall, Mask Pins, shift to use as pointer offset

  pTemp = pDrvPattern + fastTemp.word;
//  TCCR0A = fastTemp.HByte; //Disable PWM outputs (and thereby close low side FET)
//  TCCR2A = fastTemp.HByte; //Disable PWM output (and thereby close low side FET)

  PORT_MC = *(pTemp);    //Change drive levels on high side

  TCCR0A = *(pTemp + PATTERN_COM0_OFFSET);    // Reconfigure output compare operation for T0
  TCCR2A = *(pTemp + PATTERN_COM2_OFFSET); // Reconfigure output compare operation for T2
  count--;
}



//! Initialize pin change interrupts for PORTB pin 1, 2 and 3
void Init_MC_pin_change_interrupt( void )
{
  PCMSK0 = (1<<PCINT1)|(1<<PCINT2)|(1<<PCINT3); //Enable pin change interrupt on PB1:3
  PCICR = 1<<PCIE0;    // Enable pin change interrupt0 (PORTB)
}



//! Initialize motor control timers.
void Init_MC_timers( void )
{
  //Timer Counter 0. OCRA and OCRB used for motor
  TCCR0A = (1<<COM0A1)|(0<<COM0A0)|        // Clear OCRA on compare match
           (1<<COM0B1)|(0<<COM0B0)|        // Clear OCRB on compare match
           (1<<WGM01)|(1<<WGM00);         // Fast PWM mode
  TCCR0B = (0<<FOC0A)|(0<<FOC0B)|
           (0<<WGM02)|                     // Fast PWM mode
           (0<<CS02)|(0<<CS01)|(1<<CS00); // Prescaler = CLK/1

  //Timer Counter 2. OCRA and OCRB used for motor
  TCCR2A = (0<<COM2A1)|(0<<COM2A0)|        // OCRA not connected
           (1<<COM2B1)|(0<<COM2B0)|        // Clear OCRB on compare match
           (1<<WGM01)|(1<<WGM00);         // Fast PWM mode
  TCCR2B = (0<<FOC2A)|(0<<FOC2B)|
           (0<<WGM22)|                     // Fast PWM mode
           (0<<CS22)|(0<<CS21)|(1<<CS20); // Prescaler = CLK/1

  // Synchronize timers
  TCNT0 = 0;
  TCNT2 = 3;

  TIFR0 = TIFR0;    // Clear TC0 interrupt flags
  TIFR1 = TIFR1;    // Clear TC2 interrupt flags
}



//! Initialize ADC module.
void Init_ADC( void )
{
  //PRR &= ~(1 << PRADC);
  ADCSRA = (1 << ADEN) | (1 << ADPS1); // Enable ADC, clock prescaler = 4
}



//! Set motor speed, in PWM duty cycle.
void Set_Speed(unsigned char speed)
{
  TIFR0 = TIFR0;    // Clear TC0 interrupt flags
  while( !(TIFR0 & (1<<TOV0)));  // Wait for TOV to ensure that all registers are
                            //  updated in the same timer cycle
  __disable_interrupt();
  OCR0A = speed;        // Change the duty cycle
  OCR0B = speed;
  OCR2B = speed;
  __enable_interrupt();
}



//! Set motor direction, CW og CCW.
void Set_Direction(unsigned char direction)
{
  if(direction == CLOCKWISE)
  {
    __disable_interrupt();        //Variable also used in interrupt and access most be protected
    pDrvPattern = drvPatternsCW;   // Set dir to CW, by pointing to CW pattern
    __enable_interrupt();
  }
  else
  {
    __disable_interrupt();        //Variable also used in interrupt and access most be protected
    pDrvPattern = drvPatternsCCW;   // Set dir to CCW, by pointing to CCW pattern
    __enable_interrupt();
  }
}


//! Application starts here.
void main( void )
{
  unsigned char speed = 0;
  unsigned char setspeed = 0;
  signed int current;
  MCUCR |= (1<<PUD);  // Disable all pull-ups
  hallMask = HALL_MASK; // Initialize hallMask variable
  //Set initial direction.
  Set_Direction( CLOCKWISE );

  Init_MC_timers();
  Init_MC_pin_change_interrupt();
  Init_ADC();

  DDR_HALL |= HALL_MASK;    //Lock HALL sensor by driving Hall lines
  PORT_HALL |= HALL_MASK;
  PORT_HALL &= ~HALL_MASK;  //Release HALL sensor lines and trigger PC interrupt
  DDR_HALL &= ~HALL_MASK;
  __enable_interrupt();
  Set_Speed(speed);
  DDR_MC = MC_MASK;        // Enable outputs

  DDRC |= (1<<PC1);
  for(;;) {
    // Start ADC conversion, shunt high side
    ADMUX = ADC_MUX_SHUNT_H;
    ADCSRA |= (1 << ADSC);
    // Wait for ADC conversion
    do {} while (ADCSRA & (1 << ADSC));
    current = ADCH; // Shunt low side is GND.

    // If current consumption is too high, limit current
    if (current > MAX_CURRENT_ADC ) {
      PORTC &= ~(1<<PC1);
      if( speed >= 2 ) speed -= 2; // Slow down if too fast.
    } else {
      PORTC |= (1<<PC1);
      // Start ADC conversion, speed ref
      ADMUX = ADC_MUX_SPEED_REF;
      ADCSRA |= (1 << ADSC);
      // Wait for ADC conversion
      do {} while (ADCSRA & (1 << ADSC));
      setspeed = ADCH*2; // Potmeter fed from 2.5V.

      // Approach speed set point.
      if( setspeed > speed ) ++speed; else
        if( setspeed < speed ) --speed;
    }

    Set_Speed(speed);
  }
}

